
love.filesystem.load("settings.lua")()
love.filesystem.load("blueprints.lua")()
love.filesystem.load("wallmaker.lua")()
love.filesystem.load("marshaller.lua")()
love.filesystem.load("client.lua")()
love.filesystem.load("server.lua")()

walls = {
	--{ x1 = -550, y1 = -550, x2 = -500, y2 = -500, x3 =  500, y3 = -500, x4 =  550, y4 = -550 },
	--{ x1 =  550, y1 = -550, x2 =  500, y2 = -500, x3 =  500, y3 =  500, x4 =  550, y4 =  550 },
	--{ x1 =  550, y1 =  550, x2 =  500, y2 =  500, x3 = -500, y3 =  500, x4 = -550, y4 =  550 },
	--{ x1 = -550, y1 =  550, x2 = -500, y2 =  500, x3 = -500, y3 = -500, x4 = -550, y4 = -550 },
	
	{ x1 = -550, y1 = -550, x2 = -500, y2 = -500, x3 =  100, y3 = -500, x4 =  100, y4 = -550 },
	{ x1 =  550, y1 = -550, x2 =  500, y2 = -550, x3 =  500, y3 =  550, x4 =  550, y4 =  550 },
	{ x1 =  100, y1 =  550, x2 =  100, y2 =  500, x3 = -500, y3 =  500, x4 = -550, y4 =  550 },
	{ x1 = -550, y1 =  550, x2 = -500, y2 =  500, x3 = -500, y3 = -500, x4 = -550, y4 = -550 },
}

ground = {
	{ x1 = -500, y1 = -500, x2 = 500, y2 = -500, x3 = 500, y3 = 500, x4 = -500, y4 = 500 },
	
	{ x1 = 100, y1 = -550, x2 = 100, y2 = -500, x3 = 500, y3 = -500, x4 = 500, y4 = -550 },
	{ x1 = 100, y1 = 550, x2 = 100, y2 = 500, x3 = 500, y3 = 500, x4 = 500, y4 = 550 },
}


-- Starting position
startWall(vector{x = 300, y = -550}, north)

function leftTurnRoad()
	circleSteps = 16
	for i = 1, circleSteps do
		turn(-math.pi / 2 / circleSteps)
		moveForward(100)
		makeDoubleWall(200)
	end
end

leftTurnRoad()
leftTurnRoad()

moveForward(550)
makeDoubleWall(200)
moveForward(550)
makeDoubleWall(200)

leftTurnRoad()
leftTurnRoad()




blueprintBar =
{
	[1] = templates.obj1,
	[2] = templates.thruster1,
	[3] = templates.laser1,
	[4] = templates.grapplinghookgun,
	[5] = templates.torqueupgrade,
}

pixelsPerMeter = 20

customBlueprint = { }

function buildNewWorld()
	if world then
		world:destroy()
	end
	
	world = love.physics.newWorld(0, 0, true)
	
	world:setCallbacks(collisionBeginContact, nil, nil, nil)
	
	objects = {}
	
	-- Random look!-an-outta-nowhere block
	local b = { }
	b.shape = love.physics.newRectangleShape(30, 30)
	b.body = love.physics.newBody(world, 200, 200, "static")
	b.fixture = love.physics.newFixture(b.body, b.shape)
	b.fixture:setUserData(b)
	b.wall = true
	objects.block = b
	registerObject(b, -1)
	
	-- Walls
	for i, wall in ipairs(walls) do
		if not wall.shape then
			wall.coords = {wall.x1, wall.y1, wall.x2, wall.y2,
						   wall.x3, wall.y3, wall.x4, wall.y4}
			wall.shape = love.physics.newPolygonShape(unpack(wall.coords))
			wall.wall = true
			
			wall.callWhenDestroyed = function(self, func)
				-- Just ignore it... walls never get destroyed.
			end
		end
		
		wall.body = love.physics.newBody(world, 0, 0, "static")
		wall.fixture = love.physics.newFixture(wall.body, wall.shape)
		wall.fixture:setUserData(wall)
		
		registerObject(wall, -i - 1) -- -1 because of the random practice block
	end
end

function love.load()
	-- Auto-flush after every line on stdout/stderr
	io.stdout:setvbuf("line")
	io.stderr:setvbuf("line")
	
	-- Check graphics capabilities
	settings.useCanvases = settings.useCanvases and love.graphics.isSupported("canvas")
	settings.usePixelEffects = settings.usePixelEffects and love.graphics.isSupported("pixeleffect")
	
	-- Look up the functions needed by the marshaller
	generateMessageTypes()
	
	-- Start the server and client
	startServer()
	startClient()
	
	love.physics.setMeter(pixelsPerMeter)
	
	-- Create the physical level (physics world, walls, etc)
	buildNewWorld()
	
	-- Best pixels ever! Yes, an actual file isn't needed when there's ImageData...
	-- but the ridiculousness of having a 1-pixel .png file makes this simply awesome.
	bluePixel = love.graphics.newImage("resources/bluepixel.png")
	whitePixel = love.graphics.newImage("resources/whitepixel.png")
	
	-- Cache screen width and height... these are unlikely to change often
	screenWidth = love.graphics.getWidth()
	screenHeight = love.graphics.getHeight()
	
	-- Stencil to clip blueprint bar snapshots
	function barStencilFunction( )
		local boxSize = 30
		love.graphics.setColor(255, 255, 255, 255)
		love.graphics.rectangle("fill", -boxSize / 2 + 1, -boxSize / 2 + 1, boxSize - 2, boxSize - 2)
	end
	blueprintBarStencil = love.graphics.newStencil(barStencilFunction)
	
	
	-- Draw and store the many, many ground dots into a canvas for later usage
	groundTextureData = love.image.newImageData(10, 10)
	groundTextureData:setPixel(0, 0, 255, 255, 255, 255)
	groundTexture = love.graphics.newImage(groundTextureData)
	groundTexture:setWrap("repeat", "repeat")
	groundTilesQuad = love.graphics.newQuad(0, 0, 1024, 1024, 10, 10)
	
	function drawGroundDots(x1, y1, x2, y2)
		love.graphics.drawq(groundTexture, groundTilesQuad, x1, y1)
	end
	
	-- Draw the level's static objects (walls and ground)
	function drawLevelGround(red, green, blue, ignoreVisibility, drawStencil)
		local visible, notVisible = 0, 0
		
		for i, g in ipairs(ground) do
			if not g.stencil then
				local function groundStencilFunction()
					love.graphics.setColor(255, 255, 255, 255)
					love.graphics.polygon("fill", g.x1, g.y1, g.x2, g.y2, g.x3, g.y3, g.x4, g.y4)
				end
				
				g.stencil = love.graphics.newStencil(groundStencilFunction)
				g.stencilFunction = groundStencilFunction
				g.minx = math.floor(math.min(g.x1, g.x2, g.x3, g.x4) / 10) * 10
				g.miny = math.floor(math.min(g.y1, g.y2, g.y3, g.y4) / 10) * 10
				g.maxx = math.floor(math.max(g.x1, g.x2, g.x3, g.x4) / 10) * 10
				g.maxy = math.floor(math.max(g.y1, g.y2, g.y3, g.y4) / 10) * 10
			end
			
			if ignoreVisibility or rectangleIsVisible(g.minx, g.miny, g.maxx, g.maxy) then
				visible = visible + 1
				
				if drawStencil then
					-- Actually draw the stencil instead of using it (to create
					-- a single larger stencil buffer out of smaller ones)
					g.stencilFunction()
				else
			    	love.graphics.setStencil(g.stencil)
					
					love.graphics.setColor(red, green, blue)
				    --love.graphics.setColor(255, 255, 255, 255)
				    drawGroundDots(g.minx, g.miny, g.maxx, g.maxy)
				    
				    love.graphics.setStencil()
			    end
			else
				notVisible = notVisible + 1
			end
	    end
	    
	    --print("Visible:", visible, "Not visible:", notVisible)
	end
	
	if settings.useCanvases and settings.drawGround then
		levelGround = love.graphics.newCanvas(4096, 4096)
		
	    love.graphics.setCanvas(levelGround)
	    love.graphics.push()
	    love.graphics.translate(2048, 2048)
	    
	    function compositeStencil()
		    drawLevelGround(255, 255, 255, true, true)
	    end
	    
	    local humongousStencil = love.graphics.newStencil(compositeStencil)
	    love.graphics.setStencil(humongousStencil)
	    
		local bigGroundTilesQuad = love.graphics.newQuad(0, 0, 4096, 4096, 10, 10)
		
		love.graphics.drawq(groundTexture, bigGroundTilesQuad, -2048, -2048)
	    
	    love.graphics.setCanvas()
	    love.graphics.pop()
	    
		levelGround:setFilter("nearest", "nearest")
		
   		-- Replace the original function.
	    function drawLevelGround(red, green, blue)
			love.graphics.setColor(red, green, blue)
			--love.graphics.setColor(255, 255, 255)
			love.graphics.draw(levelGround, -2048, -2048)
		end
	end
	
	function drawLevelWalls()
		love.graphics.setColor(255, 255, 255)
	    
		for i, wall in ipairs(walls) do
			love.graphics.polygon("fill", unpack(wall.coords))
		end
	end
	
	if settings.useCanvases then
		levelWalls = love.graphics.newCanvas(4096, 4096)
		love.graphics.setCanvas(levelWalls)
		love.graphics.push()
		love.graphics.translate(2048, 2048)
		
		drawLevelWalls()
		
		love.graphics.pop()
		love.graphics.setCanvas()
		
   		-- Replace the original function.
		function drawLevelWalls()
			love.graphics.setColor(255, 255, 255)	
			love.graphics.draw(levelWalls, -2048, -2048)
		end
	end
	
	if settings.usePixelEffects then
		testPixelEffect = love.graphics.newPixelEffect [[
			extern float mouse_x;
			extern float mouse_y;
			extern float image_width;
			extern float image_height;
			
			vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
				float d = distance(screen_coords, vec2(mouse_x, mouse_y));
				float pixelWidth = 1 / image_width;
				float pixelHeight = 1 / image_height;
				int i;
				vec4 final_color = vec4(0, 0, 0, 0);
				
				if (d < 150) {
					float effectWeight = 1 - d / 150;
					int stretch = 5;
					
					for ( i = -stretch; i <= stretch; i++ )
						final_color += Texel(texture, vec2(texture_coords.x + i * pixelWidth, texture_coords.y));
					
					for ( i = -stretch; i <= stretch; i++ )
						final_color += Texel(texture, vec2(texture_coords.x, texture_coords.y + i * pixelHeight));
					
					final_color = final_color * effectWeight * 0.5;
				}
				
				final_color = max(final_color, Texel(texture, texture_coords));
				final_color.a = 1;
				
				return final_color * color;
			}
		]]
		
		shockwavePixelEffect = love.graphics.newPixelEffect [[
			extern float center_x;
			extern float center_y;
			extern float image_width;
			extern float image_height;
			extern float time;
			const float pi = 3.14159265358979323846264;
			
			
			vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
				vec2 pixelSize = vec2(1 / image_width, 1 / image_height);
				vec2 center = vec2(center_x, center_y) * pixelSize + 0.5;
				
				float strength = max(0, pow(1 - time / 3, 1.5));
				
				vec2 fromCenter = center - texture_coords;
				vec2 dir = normalize(fromCenter) * pixelSize;
				
				float dist = length(fromCenter / pixelSize / 100);
				float x = (dist - time) * 2 * time + 1;
				float displacement = sin((1 - x * x) * pi);
				
				if (x >= 0 && x < 1)
					return Texel(texture, texture_coords + strength * dir * displacement * 15) * color;
				else
					return Texel(texture, texture_coords) * color;
			}
		]]
		
		refractiveBubblePixelEffect = love.graphics.newPixelEffect [[
			extern float center_x;
			extern float center_y;
			extern float image_width;
			extern float image_height;
			extern float time;
			const float pi = 3.14159265358979323846264;
			
			vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
				vec2 pixelSize = vec2(1 / image_width, 1 / image_height);
				vec2 center = vec2(center_x, center_y) * pixelSize + 0.5;
				
				// Speed in pixels per second
				float speed = time * 150;
				
				// Length in pixels at which the strength completely decays.
				float maxLength = 150;
				
				// The position of the current point from the center of the bubble.
				vec2 fromCenter = texture_coords - center;
				
				// The following are inside the [0..1) interval.
				float edgeRadius = sqrt(length(speed * normalize(pixelSize)) / maxLength);
				float pointRadius = length(fromCenter / pixelSize) / maxLength;
				float nLength = pointRadius / edgeRadius;
				
				float strength = max(0, pow(1 - edgeRadius, 2));
				
				if (nLength < 1 && edgeRadius < 1)
				{
					texture_coords += fromCenter * strength * (pow(nLength, 3) - 1);
					
					color.rg += 0.5 * strength;
					if (color.r > 1 || color.g > 1)
						color.rg = vec2(1, 1);
				}
				
				return Texel(texture, texture_coords) * color;
			}
		]]
		
		local warnings = ""
		warnings = warnings .. " testPixelEffect: " .. testPixelEffect:getWarnings()
		warnings = warnings .. " shockwavePixelEffect: " .. shockwavePixelEffect:getWarnings()
		warnings = warnings .. " refractiveBubblePixelEffect: " .. refractiveBubblePixelEffect:getWarnings()
		
		print("Pixel effect messages:")
		print(warnings)
	else
		testPixelEffect = nil
		shockwavePixelEffect = nil
		refractiveBubblePixelEffect = nil
	end
end

function rectangleIsVisible(minx, miny, maxx, maxy)
	local cx, cy = getCameraOffsets()

	if maxx - cx < 0 or maxy - cy < 0
			or minx - cx > screenWidth or miny - cy > screenHeight then
		return false
	end
	
	return true
end

function notifyWhenCollected(t, name)
	local mt = {
		__gc = function()
			print("Table " .. name .. " (" .. tostring(t) .. ")" ..
					" has been collected.")
		end
	}
	
	setmetatable(t, mt)
end

function makePlayerVehicle(playerName)
	-- Player vehicle
	local v = registerObject(templates.vehicle.build(0, 0, 0))
	
	if playerName == myName then
		objects.vehicle = v
	end
	
	v.player = playerName
	
	return v
end

collisions = { }

function collisionBeginContact(f1, f2, contact)
	local obj1, obj2 = f1:getUserData(), f2:getUserData()
	
	-- They need to be ordered to prevent a desync... so table.insert is -required-.
	
	if not obj1.wall and obj1.blueprint.collision then
		table.insert(collisions, {obj1, obj2})
	end
	
	if not obj2.wall and obj2.blueprint.collision then
		table.insert(collisions, {obj2, obj1})
	end
end

function handleCollisions()
	for i, v in ipairs(collisions) do
		local obj1, obj2, contact = unpack(v)
		obj1.blueprint.collision(obj1, obj2, contact)
	end
	
	collisions = { }
end

function getBlueprintBoundingBox(customBlueprint, x, y, r)
	local topLeftX, topLeftY = 0
	local bottomRightX, bottomRightY = 0
	
	for i, bpObj in ipairs(customBlueprint) do
		
		-- ...?
	end
end

function storeCustomBlueprint()
	if not designingCustomBlueprint then
		return activeBlueprint
	end
	
	local newBlueprint = customBlueprint
	customBlueprint = { }
	designingCustomBlueprint = nil
	
	-- Change all objects to be relative to the center of this custom blueprint.
	local cr = objects.vehicle.body:getAngle()
	
	for i, bpObj in ipairs(newBlueprint) do
		bpObj.x, bpObj.y = objects.vehicle.body:getLocalPoint(bpObj.x, bpObj.y)
		bpObj.x = bpObj.x - 100
		bpObj.r = bpObj.r - cr
	end
	
	function customDraw()
		for i, bpObj in ipairs(newBlueprint) do
			love.graphics.push()
			love.graphics.translate(bpObj.x, bpObj.y)
			love.graphics.rotate(bpObj.r)
			bpObj.blueprint.draw()
			love.graphics.pop()
		end
	end
	
	function customDrawShadow()
		for i, bpObj in ipairs(newBlueprint) do
			love.graphics.push()
			love.graphics.translate(bpObj.x, bpObj.y)
			love.graphics.rotate(bpObj.r)
			bpObj.blueprint.drawShadow()
			love.graphics.pop()
		end
	end
	
	newBlueprint.draw = customDraw
	newBlueprint.drawShadow = customDrawShadow
	newBlueprint.build = customBuild
	newBlueprint.isCustomBlueprint = true
	newBlueprint.snapshotScale = 0.25
	
	return newBlueprint
end

function sendMaterializeCustomBlueprint(customBlueprint, cx, cy, cr)
	-- Send stuff.
	for i, bpObj in ipairs(customBlueprint) do
		if not bpObj.sentToServer then
			-- Declare that we're putting stuff in a custom blueprint.
    		if not sentBlueprintQueueCommand then
    			--sendCall.startCustomBlueprint()
    		end
    		
    		if designingCustomBlueprint then
				-- Add a blueprint to another blueprint
				if bpObj.toObj then
					guiMaterializeAndConnectObject(bpObj.blueprint, bpObj.toObj,
							bpObj.hardpoint, bpObj.hardpointNr, bpObj.connector, bpObj.connectorNr)
				else
					local x, y = localToWorld(cx, cy, cr, bpObj.x, bpObj.y)
					local r = cr + bpObj.r
					
					guiMaterializeObject(bpObj.blueprint, x, y, r)
				end
			else
	    		-- The signatures of these messages will make sure it
	    		-- takes only the right fields from the bpObj table.
	    		if bpObj.toObj then
					sendToServer("addConnectedObjectToBlueprint", bpObj)
				else
					sendToServer("addObjectToBlueprint", bpObj)
				end
			end
		end
	end
	
	sendCall.materializeCustomBlueprint(cx, cy, cr)
end

function love.keypressed(key, unicode)
    if key == 'q' or key == "escape" then
    	love.event.quit()
   	end
   	
    -- A number key from 1 to 9 was pressed?
    local nk = string.byte(key)
    local n0 = string.byte('0')
    
    if #key == 1 and nk >= n0 + 1 and nk <= n0 + 9 then
    	-- ToDo: By the time this function is processed, modifiers may change.
    	-- Nothing that can be done about it... but keep that in mind for when
    	-- a port to a more capable system happens.
    	-- ToDo: maybe consider remembering Shift/Ctrl press events instead?
    	
    	if love.keyboard.isDown("lctrl", "rctrl") then
    		blueprintBarOld[nk - n0] = blueprintBar[nk - n0]
    		blueprintBarOldGuiTransition[nk - n0] = 0
    		
    		blueprintBar[nk - n0] = storeCustomBlueprint()
    		activeBlueprint = nil
    	else
	   		activeBlueprint = blueprintBar[nk - n0]
	   	end
   		
   		if activeBlueprint then
			activeBlueprintButton = nk - n0
			activeBlueprintGuiTransition = 0
			currentConnector = 1
		end
    end
    
    -- 0 just deselects
    if key == "0" then
    	activeBlueprint = nil
    end
    
    if objects.vehicle then
	    if key == "w" or key == "up" then
	    	sendCall.forward()
	    elseif key == "s" or key == "down" then
	    	sendCall.backward()
	    elseif key == "a" or key == "left" then
	    	sendCall.left()
	    elseif key == "d" or key == "right" then
	    	sendCall.right()
	    end
    	
	    if key == "e" then
	    	sendCall.fireVehicleHook()
	    end
    	
	    if key == "f" then
	    	sendCall.radialSignal()
	    end
	    
		if key == "b" then
			if not designingCustomBlueprint then
				designingCustomBlueprint = true
			else
				designingCustomBlueprint = false
				
				sendMaterializeCustomBlueprint(customBlueprint, 0, 0, 0)
				customBlueprint = { }
			end
		end
    end
    
    if key == "o" then
    	sendToServerRaw("state-broadcast - [" .. saveGameState() .. "]")
    end
    
    -- Latency control
    if key == "+" or key == "kp+" then
    	if latency < 0.03 then
    		latency = 0
    	end
    	
    	latency = latency + 0.05
    elseif key == "-" or key == "kp-" then
    	latency = latency - 0.05
    	
    	if latency < 0.02 then
    		latency = 0.02
    	end
    end
    
    -- The 'somehow modify random thing' button.
    if key == "t" then
    	testingStuff = not testingStuff
    end
end

function love.keyreleased(key, unicode)
	if objects.vehicle then
		if key == "w" or key == "up" then
			sendCall.forwardStop()
		elseif key == "s" or key == "down" then
			sendCall.backwardStop()
		elseif key == "a" or key == "left" then
			sendCall.leftStop()
		elseif key == "d" or key == "right" then
			sendCall.rightStop()
		end
	end
end

-- Movement commands
function forward(vehicle) vehicle.forward = true end
function backward(vehicle) vehicle.backward = true end
function left(vehicle) vehicle.left = true end
function right(vehicle) vehicle.right = true end

function forwardStop(vehicle) vehicle.forward = false end
function backwardStop(vehicle) vehicle.backward = false end
function leftStop(vehicle) vehicle.left = false end
function rightStop(vehicle) vehicle.right = false end

function guideAngle(vehicle, r)
	vehicle.guidingAngle = r
end

function stopGuiding(vehicle)
	vehicle.guidingAngle = false
end

override = false

function updatePhysics(dt)
	-- Laser beams - not a force, but instant beams last as long as a physics tick
	for beam in pairs(laserBeams) do
		if beam.instant and not beam.life then
			beam.life = 0
		end
	end
	
	-- Player vehicles
	-- ToDo: this is pairs(), not ipairs... I hope simply applying forces in
	-- an unpredictable order will not be a cause for desynchronization
	for player, vehicle in pairs(players) do
		local r = vehicle.body:getAngle()
		local torque = 0
		
		if (override and love.keyboard.isDown("i")) or vehicle.forward then
			--[[
			vehicle.body:applyForce(math.cos(r)*500, math.sin(r)*500,
					vehicle.body:getWorldPoint(0, -5))
			vehicle.body:applyForce(math.cos(r)*500, math.sin(r)*500,
					vehicle.body:getWorldPoint(0, 5))
			--]]
			
			vehicle.body:applyForce(math.cos(r)*1000, math.sin(r)*1000)
		end
		
		if (override and love.keyboard.isDown("k")) or vehicle.backward then
			vehicle.body:applyForce(-math.cos(r)*1000, -math.sin(r)*1000)
		end
		
		if (override and love.keyboard.isDown("l")) or vehicle.right then
			--[[
			vehicle.body:applyForce(math.cos(r)*500, math.sin(r)*500,
					vehicle.body:getWorldPoint(0, -5))
			vehicle.body:applyForce(-math.cos(r)*500,-math.sin(r)*500,
					vehicle.body:getWorldPoint(0, 5))
			--]]
			
			torque = torque + 2000
		end
		
		if (override and love.keyboard.isDown("j")) or vehicle.left then
			torque = torque - 2000
		end
		
		-- Guiding angle with mouse
		if vehicle.guidingAngle then
			--local x, y = vehicle.body:getLocalPoint(vehicle.guidingX, vehicle.guidingY)
			--local guidingAngle = math.atan2(y, x) / math.pi
			local guidingAngle = vehicle.guidingAngle - vehicle.body:getAngle()
			
			guidingAngle = (guidingAngle + math.pi) % (2*math.pi) - math.pi
			
			torque = torque + 4000 * (guidingAngle / math.pi) * settings.guidingSensitivity
		end
		
		torque = minMax(torque, -4000, 4000)
		vehicle.body:applyTorque(torque)
		
		-- Would be nice if the torque could be applied at the hardpoint's
		-- location... but Box2D does not (yet) support that.
		-- ToDo: maybe replace with a couple applyForce()s?
		for i, topHp in ipairs(vehicle.blueprint.topHardpoints or { }) do
			local module = vehicle.connectedObjects[topHp]
			if module and module.blueprint == templates.torqueupgrade then
				vehicle.body:applyTorque(torque * 3)
			end
		end
	end
	
	-- Updates forces on dynamic objects
	
	-- ToDo: ugly hack to process them in ID order... need to find a better way
	local sortedObjects = {}
	for obj in pairs(dynamicObjects) do
		if obj.blueprint.updatePhysics then
			table.insert(sortedObjects, obj)
		end
	end
	table.sort(sortedObjects, function (a, b) return a.id < b.id end)
	for i, obj in ipairs(sortedObjects) do
		if not obj.destroyed then
			obj.blueprint.updatePhysics(obj, dt)
		end
	end
	--[[
	for obj in pairs(dynamicObjects) do
		if obj.blueprint.updateForces then
			obj.blueprint.updateForces(obj)
		end
	end
	--]]
end

local time = 0
local timeStep = 1 / 60

local localTime = 0

updates = 0

local lastGuideMessage = 0

function love.update(dt)
	localTime = localTime + dt

	-- Receive, handle and send network messages
	checkServer()
	checkClient()
	
	-- Guide angle with mouse
	if love.mouse.isDown("r") and objects.vehicle then
		local updateRate = 1 / 10
		
		guiding = true
		
		if lastGuideMessage > 0 then
			lastGuideMessage = lastGuideMessage - dt
		end
		
		local mx, my = getCameraToWorldPoint(love.mouse.getPosition())
		--local x, y = objects.vehicle.body:getLocalPoint(mx, my)
		local vx, vy = objects.vehicle.body:getPosition()
		
		r1 = math.atan2(my - vy, mx - vx)
		x1 = mx
		y1 = my
		
		if lastGuideMessage <= 0 then
			-- Since (for now) it's sent in textual form, make it have a few
			-- less digits to make it more network-friendly.
			sendCall.guideAngle(math.floor(r1 * 1000) / 1000)
			
			lastGuideMessage = updateRate
		end
	elseif guiding then
		sendCall.stopGuiding()
		guiding = false
	end
	
    -- Start CPU usage timer (physics)
    local physicsTime1 = love.timer.getMicroTime()

--[[
	time = time + dt
	while time > timeStep do
		updateForces()
	
		world:update(timeStep)
		time = time - timeStep
	end
--]]
	
	while processTickMessages() do
		updatePhysics(timeStep)
		
		world:update(timeStep)
		handleCollisions()
		
		updates = updates + 1
	end

	-- Above this, update forces; below, update only positions
	
	-- Where an object would be if welded to another object
	local function transformToWeldedPosition(x, y, r, hardpoint, connector)
		local hx, hy = unpack(hardpoint)
		local cx, cy = unpack(connector)
		
		x, y = localToWorld(x, y, r, hx, hy)
		r = r + hardpoint.angle - connector.angle + math.pi
		x, y = localToWorld(x, y, r, -cx, -cy)
		
		return x, y, r
	end
	
	-- Update blueprint objects
	if designingCustomBlueprint then
		for i, bpObj in ipairs(customBlueprint) do
			if bpObj.toObj then
				local x, y, r
				if bpObj.toObj.partOfBlueprint then
					x = bpObj.toObj.x
					y = bpObj.toObj.y
					r = bpObj.toObj.r
				else
					x, y = bpObj.toObj.body:getPosition()
					r = bpObj.toObj.body:getAngle()
				end
				
				bpObj.x, bpObj.y, bpObj.r = transformToWeldedPosition(x, y, r,
						bpObj.hardpoint, bpObj.connector)
			end
		end
	end
	
	-- Update camera offsets
	if objects.vehicle then
		local x, y = objects.vehicle.body:getPosition()
		
		function cameraTransform1(x, y)
			return x / 1.5, y / 1.5
		end
		
		function cameraTransform2(x, y)
			return x, y
		end
		
		function cameraTransform3(x, y)
			return x / 1.05, y / 1.05
		end
		
		cameraFunctions = cameraFunctions or {
			[1] = { f = cameraTransform1, activeWeight = 0.8 },
			[2] = { f = cameraTransform2, activeWeight = 0.1 },
			[3] = { f = cameraTransform1, activeWeight = 0.1 },
		}
		
		-- Transition camera to..
		if x > -500 and y > -500 and x < 500 and y < 500 then
			transitionToCamera = 1
		else
			transitionToCamera = 2
		end
		
		camera.x = 0
		camera.y = 0
		
		local transitionSpeed = minMax(dt * 6, 0, 1)
		
		for i, cam in ipairs(cameraFunctions) do
			if i == transitionToCamera then
				cam.activeWeight = cam.activeWeight + (1 - cam.activeWeight) * transitionSpeed
			else
				cam.activeWeight = cam.activeWeight * (1 - transitionSpeed)
			end
			
			local cx, cy = cam.f(x, y)
			camera.x = camera.x + cx * cam.activeWeight
			camera.y = camera.y + cy * cam.activeWeight
		end
	else
		local scrollSpeed = 200
		
		-- Spectator? Then use w/a/s/d keys to control the camera instead
		if love.keyboard.isDown("w") then
			camera.y = camera.y - scrollSpeed * dt
		end
		if love.keyboard.isDown("s") then
			camera.y = camera.y + scrollSpeed * dt
		end
		if love.keyboard.isDown("a") then
			camera.x = camera.x - scrollSpeed * dt
		end
		if love.keyboard.isDown("d") then
			camera.x = camera.x + scrollSpeed * dt
		end
	end
	
	-- Laser beams
	for beam in pairs(laserBeams) do
		if beam.life then
			beam.life = beam.life - dt
			
			if beam.life < 0 then
				laserBeams[beam] = nil
			end
		end
	end
	
	-- Particle systems
	for system in pairs(ephemeralParticleSystems) do
		system:update(dt)
		
		if system:isEmpty() and not system:isActive() then
			ephemeralParticleSystems[system] = nil
		end
	end
	
	-- Updates positions/particles/etc on dynamic objects
	for obj in pairs(dynamicObjects) do
		if obj.blueprint.updatePositions then
			obj.blueprint.updatePositions(obj, dt)
		end
	end
	
	-- Pixel effects
	if testingStuff then
		effectTime = 0
	elseif not effectTime then
		effectTime = 0
	else
		effectTime = effectTime + dt;
	end
	
	-- GUI
	activeBlueprintGuiTransition = minMax(activeBlueprintGuiTransition +
								4 * dt * (activeBlueprint and 1 or -1), 0, 1)
	selectedObjectGuiTransition = minMax(selectedObjectGuiTransition +
								4 * dt, 0, 1)
	guiDesyncNotice = minMax(guiDesyncNotice - dt, 0, 1)
	-- minMax it because floating point is horrible
	globalGuiSineTransition = minMax(math.sin(localTime * 2*math.pi) / 2 + 0.5, 0, 1)
	
	if keepSelectedObjectVisible == 1 and globalGuiSineTransition > 0.98 then
		keepSelectedObjectVisible = 0.98
	end
	
	if keepSelectedObjectVisible < 1 then
		keepSelectedObjectVisible = minMax(keepSelectedObjectVisible - dt, 0, 1)
	end
	
	if selectedObjectGuiLast and selectedObjectGuiTransition == 1 then
		selectedObjectGuiLast = nil
	end
	
	for i = 1, 9 do
		if blueprintBarOldGuiTransition[i] then
			blueprintBarOldGuiTransition[i] = blueprintBarOldGuiTransition[i] + dt * 6
			if blueprintBarOldGuiTransition[i] >= 1 then
				blueprintBarOldGuiTransition[i] = nil
			end
		end
	end
	
    -- CPU usage timer (physics)
    local now = love.timer.getMicroTime()
    table.insert(cpuUsagePhysics, {time = now, delta = now - physicsTime1, updates = updates})
    updates = 0
    
    while cpuUsagePhysics[1].time < now - 1 do
    	table.remove(cpuUsagePhysics, 1)
    end
end

function localToWorld(x, y, r, x2, y2)
	local sinr, cosr = math.sin(r), math.cos(r)
	
	return x + x2 * cosr + y2 * -sinr,
	       y + x2 * sinr + y2 * cosr
end

function minMax(value, min, max)
	return math.min(math.max(value, min), max)
end

-- Get world coordinates from mouse coordinates by reversing the camera offset
function getCameraToWorldPoint(x, y)
	return x + camera.x - screenWidth / 2,
	       y + camera.y - screenHeight / 2
end

function getCameraOffsets()
	return camera.x - screenWidth / 2,
	       camera.y - screenHeight / 2
end

camera = { x = 0, y = 0 }

laserBeams = {}

guiDesyncNotice = 0

ephemeralParticleSystems = {}

cpuUsageGraphics = { }
cpuUsagePhysics = { }

function printDebugStuff()
	local debugMessages = { }
	local offset = 10
	
	function printDebug(str)
	    love.graphics.print(str, 10, offset)
	    offset = offset + 20
	end
	
	love.graphics.setColor(128, 138, 128, 255)
    printDebug("Hello World")
    if r1 then
		printDebug("r1 = " .. r1)
	end
	if x1 and y1 then
    	printDebug("x1 = " .. x1 .. ", y1 = " .. y1)
    end
    printDebug("fps = " .. love.timer.getFPS())
    
    -- Calculate CPU usage totals
    local totalGraphics = 0
    local totalPhysics = 0
    local totalUpdates = 0
    for i, v in ipairs(cpuUsageGraphics) do
    	totalGraphics = totalGraphics + v.delta
    end
    for i, v in ipairs(cpuUsagePhysics) do
    	totalPhysics = totalPhysics + v.delta
    	totalUpdates = totalUpdates + v.updates
    end
    
    printDebug(string.format("cpu usage (graphics) = %.3f (%d deltas)", totalGraphics, #cpuUsageGraphics))
    printDebug(string.format("cpu usage (physics) = %.3f (%d deltas)", totalPhysics, #cpuUsagePhysics))
    printDebug(string.format("update ticks = %d (%d deltas)", totalUpdates, #cpuUsagePhysics))
    
    if testingStuff ~= nil then
    	printDebug("testing-related code blocks " .. (testingStuff and "enabled" or "disabled"))
    end
    
    if cameraFunctions and testingStuff then
    	for i, cam in ipairs(cameraFunctions) do
    		printDebug(string.format("camera #%d weight: %.3f", i, cam.activeWeight))
    	end
    end
    
    local starved = (nextTick - currentTick) == 0
    printDebug("buffered ticks: " .. string.format("%02d", nextTick - currentTick) ..
    		" (latency: " .. latency .. "s)" ..
    		(starved and " (starved! increase latency with '+' key)" or ""))
    
    if showMaxLagSpike then
    	printDebug(string.format("longest lag spike: %.3fs", showMaxLagSpike))
    end
    
    if ping then
    	printDebug(string.format("ping: %.3fs", ping))
    end
    
    if effectiveCommandLatency then
    	printDebug(string.format("effective latency: %.3fs", effectiveCommandLatency))
    end
    
    if newestTickReport then
    	for name, t in pairs(newestTickReport) do
    		if name ~= myName then
    			printDebug("desync report [" .. name .. "]: " .. (t and t.diff or "?"))
    		end
    	end
    end
    
    if not connected() then
    	printDebug("oops... connection to server is gone")
    end
    
    if designingCustomBlueprint then
    	printDebug("designing blueprint mode enabled")
    end
    
    if guiDesyncNotice > 0 then
    	printDebug("State broadcast received; reloading state!")
    end
    
    -- Speed in km/h
    if objects.vehicle then
    	local vx, vy = objects.vehicle.body:getLinearVelocity()
    	local speed = math.sqrt(vx*vx + vy*vy) / pixelsPerMeter
    	local kph = math.floor(speed * 3.6)
    	local spin = math.floor(math.abs(objects.vehicle.body:getAngularVelocity() / math.pi))
    	
    	if kph > 40 then
    		printDebug("speed = " .. kph .. "km/h")
    	end
    	
    	if spin > 5 then
    		printDebug("I.. feel.. so dizzy... (" .. spin .. "r/s)")
    	end
    end
end


activeBlueprint = nil
activeBlueprintButton = 1
activeBlueprintGuiTransition = 0
blueprintBarOld = { }
blueprintBarOldGuiTransition = { }
currentConnector = 1

selectedObject = nil
selectedObjectGuiTransition = 0
selectedObjectGuiLast = nil
keepSelectedObjectVisible = 0

globalGuiSineTransition = 0


function love.draw()
    -- Start CPU usage timer (graphics)
    local graphicsTime1 = love.timer.getMicroTime()
    
    -- Set camera position (pixel-perfect position)
    local cx, cy = getCameraOffsets()
    love.graphics.push()
    love.graphics.translate(math.floor(-cx), math.floor(-cy))
    
    -- Ground dots
    local red = 128 + 127*guiDesyncNotice*2
    local greenBlue = 128 -- 64*guiDesyncNotice*2
    
    if settings.drawGround then
    	local x, y = getCameraToWorldPoint(love.mouse.getPosition())
    	
	    love.graphics.setScissor(0, 0, love.graphics.getWidth(), love.graphics.getHeight())
	    if settings.usePixelEffects then
			refractiveBubblePixelEffect:send("center_x", x)
			refractiveBubblePixelEffect:send("center_y", 1 - y)
			refractiveBubblePixelEffect:send("image_width", 4096)
			refractiveBubblePixelEffect:send("image_height", 4096)
			refractiveBubblePixelEffect:send("time", effectTime)
			love.graphics.setPixelEffect(refractiveBubblePixelEffect)
		end
		
	    drawLevelGround(red, greenBlue, greenBlue)
	    
		if settings.usePixelEffects then
			love.graphics.setPixelEffect()
		end
		love.graphics.setScissor()
    end
	
    love.graphics.pop()
    
    -- Set camera position again (smooth)
    love.graphics.push()
    love.graphics.translate(-cx, -cy)
	
	-- Particle systems from dynamic objects
	for obj in pairs(dynamicObjects) do
		if obj.blueprint.drawParticleSystems then
			obj.blueprint.drawParticleSystems(obj)
		end
	end
	
    -- Practice block
    love.graphics.setColor(255, 255, 255)
    love.graphics.polygon("fill", objects.block.body:getWorldPoints(objects.block.shape:getPoints()))
    
    -- Objects (including player vehicle)
    for obj in pairs(dynamicObjects) do
    	if obj.body then
	    	love.graphics.push()
	    	love.graphics.translate(obj.body:getPosition())
	    	love.graphics.rotate(obj.body:getAngle())
	    	
	    	obj:draw()
	    	
	    	-- Things on top of it
	    	
		    for topHardpoint, connectedObj in pairs(obj.connectedObjects) do
				if connectedObj.blueprint.mountOn then
					love.graphics.push()
					love.graphics.translate(unpack(topHardpoint))
					love.graphics.rotate(topHardpoint.angle)
					connectedObj.blueprint.draw(connectedObj)
					love.graphics.pop()
				end
			end
	    	
	    	love.graphics.pop()
	    end
    end
    
    -- Laser beams
    for beam in pairs(laserBeams) do
    	local blueAlpha = 255 * (beam.instant and 1 or (beam.life / beam.maxLife))
    	love.graphics.setColor(255, 0, blueAlpha, blueAlpha)
    	love.graphics.line(unpack(beam.coords))
    end
    
    -- Walls
    love.graphics.setScissor(0, 0, love.graphics.getWidth(), love.graphics.getHeight())
    drawLevelWalls()
    love.graphics.setScissor()
    
    -- Sparks
    for system in pairs(ephemeralParticleSystems) do
	    love.graphics.draw(system)
	end
	
    -- Objects on the current custom blueprint (design mode only)
    if designingCustomBlueprint then
    	for i, bpObj in ipairs(customBlueprint) do
    		love.graphics.push()
    		love.graphics.translate(bpObj.x, bpObj.y)
    		love.graphics.rotate(bpObj.r)
    		bpObj.blueprint.drawShadow()
    		love.graphics.pop()
    	end
    	
    	if objects.vehicle then
	    	love.graphics.setColor(0, 255, 0)
	    	love.graphics.setPoint(3, "smooth")
	    	love.graphics.point(objects.vehicle.body:getWorldPoints(100, 0))
    	end
    end
    
	-- Shadow of object-to-be-built
	-- Only if mouse is near enough the player
	local mx, my = getCameraToWorldPoint(love.mouse.getPosition())
	if activeBlueprint then
		-- Check for the nearest hard-point to connect to, if any
		local obj, hardpoint = findNearestHardpoint(mx, my)
		local obj2, topHp = findNearestTopHardpoint(mx, my)
		
		if activeBlueprint.build and activeBlueprint.connectors and hardpoint then
			local angle, hx, hy
			
			if obj.partOfBlueprint then
				angle = hardpoint.angle + obj.r
				hx, hy = localToWorld(obj.x, obj.y, obj.r, unpack(hardpoint))
			else
				angle = hardpoint.angle + obj.body:getAngle()
				hx, hy = obj.body:getWorldPoint(unpack(hardpoint))
			end
			
			local connector = activeBlueprint.connectors[currentConnector]
			local cx, cy = unpack(connector)
			
			love.graphics.push()
			love.graphics.translate(hx, hy)
			love.graphics.rotate(angle)
			love.graphics.rotate(-connector.angle - math.pi)
			love.graphics.translate(-cx, -cy)
			activeBlueprint.drawShadow()
			love.graphics.pop()
		elseif activeBlueprint.mountOn and topHp and
				distanceFromVehicle(obj2.body:getWorldPoint(unpack(topHp))) < 200 then
			local angle = hardpoint.angle + obj2.body:getAngle()
			local hx, hy = obj2.body:getWorldPoint(unpack(topHp))
			
			love.graphics.push()
			love.graphics.translate(hx, hy)
			love.graphics.rotate(angle)
			activeBlueprint.drawShadow()
			love.graphics.pop()
		elseif distanceFromVehicle(mx, my) < 200 then
			love.graphics.push()
			love.graphics.translate(mx, my)
			love.graphics.rotate(objects.vehicle.body:getAngle())
			activeBlueprint.drawShadow()
			love.graphics.pop()
		end
	end
	
	-- Indicator of object-to-be-deleted (if 'r' key is held down)
	if love.keyboard.isDown("r") and objects.vehicle then
		local mx, my = getCameraToWorldPoint(love.mouse.getPosition())
		
		local obj, distance = findNearestObject(mx, my, 50,
				function(obj) return obj.blueprint ~= templates.vehicle end, true)
		
		if obj and obj.blueprint.destroy and 
				distanceFromVehicle(obj:getPosition()) < 300 then
			love.graphics.push()
			love.graphics.translate(obj:getPosition())
			love.graphics.setColor(255, 0, 0, 255)
			love.graphics.circle("line", 0, 0, 10)
			love.graphics.line(-5, -5, 5, 5)
			love.graphics.line(5, -5, -5, 5)
			love.graphics.pop()
		end
	else
		-- Show which object would be selected on left click
		local obj, dist = findNearestObject(mx, my, 30, function (obj)
			return not obj.mountedOn
		end)
		
		if obj and obj ~= selectedObject then
			local greenBlue = globalGuiSineTransition * 255
			
		    love.graphics.setColor(0, greenBlue, greenBlue)
		    love.graphics.setPoint(3, "smooth")
			love.graphics.point(obj:getPosition())
		end
	end
	
	-- Targetting reticule (if right-mouse-button is held down)
	if love.mouse.isDown("r") and objects.vehicle then
		local mx, my = getCameraToWorldPoint(love.mouse.getPosition())
		local d = distanceFromVehicle(mx, my)
		local x, y = objects.vehicle.body:getWorldPoint(d, 0)
		
		love.graphics.push()
		love.graphics.translate(x, y)
		love.graphics.rotate(objects.vehicle.body:getAngle())
		love.graphics.setColor(255, 0, 0, 192)
		love.graphics.line(-3, -2, 3, -2)
		love.graphics.line(-3, 2, 3, 2)
		love.graphics.pop()
	end
    
    love.graphics.pop() -- camera
    
    drawGui()
    
    printDebugStuff()
    
    -- CPU usage timer (graphics)
    local now = love.timer.getMicroTime()
    table.insert(cpuUsageGraphics, {time = now, delta = now - graphicsTime1})
    
    while cpuUsageGraphics[1].time < now - 1 do
    	table.remove(cpuUsageGraphics, 1)
    end
end

function findNearestHardpoint(x, y)
	local nearestObject, nearestHardpoint, nearestHardpointNr
	local nearest = 30
	local nx, ny
	
	local function checkObject(obj)
		if obj.blueprint.sideHardpoints then
			for i, hardpoint in ipairs(obj.blueprint.sideHardpoints) do
				local hx, hy
				if obj.partOfBlueprint then
					hx, hy = localToWorld(obj.x, obj.y, obj.r, unpack(hardpoint))
				else
					hx, hy = obj.body:getWorldPoint(unpack(hardpoint))
				end
				
				local dist = distance(x, y, hx, hy)
				
				if dist < 30 and dist < nearest and distanceFromVehicle(hx, hy) < 200 then
					nx, ny = hx, hy
					nearest = dist
					nearestObject = obj
					nearestHardpoint = hardpoint
					nearestHardpointNr = i
				end
			end
		end
	end
	
	for obj in pairs(dynamicObjects) do
		checkObject(obj)
	end
	
	if designingCustomBlueprint then
		for i, bpObj in ipairs(customBlueprint) do
			checkObject(bpObj)
		end
	end
	
	return nearestObject, nearestHardpoint, nearestHardpointNr
end

function findNearestTopHardpoint(x, y)
	local nearestObject, nearestTopHp, nearestTopHpNr
	local nearest = 30
	local nx, ny
	
	for obj in pairs(dynamicObjects) do
		if obj.blueprint.topHardpoints then
			for i, topHp in ipairs(obj.blueprint.topHardpoints) do
				local hx, hy = obj.body:getWorldPoint(unpack(topHp))
				local dist = distance(x, y, hx, hy)
				
				if dist < 30 and dist < nearest then
					nx, ny = hx, hy
					nearest = dist
					nearestObject = obj
					nearestTopHp = topHp
					nearestTopHpNr = i
				end
			end
		end
	end
	
	return nearestObject, nearestTopHp, nearestTopHpNr
end

function findNearestObject(x, y, maxDistance, filterFunc, includeBlueprintObjects)
	local nearestObject
	local nearestDistance
	
	local function checkObject(obj)
		if not filterFunc or filterFunc(obj) then
			local dist = distance(x, y, obj:getPosition())
			if (not nearestDistance or dist < nearestDistance) and
					(not maxDistance or dist < maxDistance) then
				nearestDistance = dist
				nearestObject = obj
			end
		end
	end
	
	if includeBlueprintObjects then
		for i, bpObj in ipairs(customBlueprint) do
			checkObject(bpObj)
		end
	end
	
	for obj in pairs(dynamicObjects) do
		checkObject(obj)
	end
	
	return nearestObject, nearestDistance
end

function distance(x1, y1, x2, y2)
	return math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2))
end

function distanceFromVehicle(x, y)
	return distance(x, y, objects.vehicle.body:getPosition())
end

function drawGui( )
	local boxSize = 30
	local width, height = love.graphics.getWidth(), love.graphics.getHeight()
	
	-- Control groups bar
	for i = 1, 9 do
		local mx, my = love.mouse.getPosition()
		local x, y = 50 + (i-1) * (boxSize + 10), height - 50 - boxSize
		local hover = false
		
		if mx >= x and mx <= x + boxSize and my >= y and my <= y + boxSize then
			hover = true
		end
		
		local color = { 128, 128, 128, 0 }
		
		if not blueprintBar[i] then
			color = { 0, 0, 0, 0 }
		elseif blueprintBar[i].isCustomBlueprint then
			color = { 128, 64, 192, 0 }
		end
		
		color[4] = hover and 255 or 128
		
		love.graphics.setColor(color)
		love.graphics.rectangle("fill", x, y, boxSize, boxSize)
		
		if hover then
			love.graphics.setColor(255, 255, 255, 255)
		else
			love.graphics.setColor(128, 128, 128, 255)
		end
		love.graphics.rectangle("line", x, y, boxSize, boxSize)
		
		if activeBlueprintButton == i and activeBlueprintGuiTransition > 0 then
			local t = activeBlueprintGuiTransition
			
			love.graphics.setColor(255, 0, 0, 192)
			love.graphics.rectangle("line", x - t * 3, y - t * 3, boxSize + t * 6, boxSize + t * 6)
		end
		
		local function drawMiniatureObject(blueprintBarObj)
			if not blueprintBarObj then
				return
			end
			
			-- This will serve as a small snapshot/thumbnail for the button
			love.graphics.push()
			love.graphics.translate(x + boxSize / 2, y + boxSize / 2)
			love.graphics.setStencil(blueprintBarStencil)
			love.graphics.rotate(-math.pi / 4)
			love.graphics.scale(blueprintBarObj.snapshotScale)
			blueprintBarObj.draw(nil)
			love.graphics.pop()
			love.graphics.setStencil()
		end
		
		local bpbTrans = blueprintBarOldGuiTransition[i]
		local drawCurrent = true
		local drawWhite = false
		
		-- Transition from old object, if any
		if bpbTrans then
			if bpbTrans < 0.5 then
				drawMiniatureObject(blueprintBarOld[i])
				
				drawWhite = 255 * bpbTrans * 2
				drawCurrent = false
			else
				drawWhite = 255 + 255 * (1 - bpbTrans * 2)
			end
		end
		
		if drawCurrent then
			drawMiniatureObject(blueprintBar[i])
		end
		
		if drawWhite then
			love.graphics.setColor(255, 255, 255, drawWhite)
			love.graphics.rectangle("fill", x, y, boxSize, boxSize)
		end
	end
	
    local cx, cy = getCameraOffsets()
	
	local function showObjectInfo(object, transition)
	    local x, y = object:getPosition()
	    
	    local circleSize = object.body and 30 or 15
	    
		love.graphics.setColor(0, 255 * transition, 255 * (1 - transition))
		love.graphics.circle("line", x - cx, y - cy, circleSize * transition)
		
		local function stencilFunc()
			love.graphics.setColor(255, 255, 255)
			love.graphics.circle("fill", x - cx, y - cy, circleSize * transition)
		end
		
		-- And some information about this object
		local function interpolate(x1, y1, x2, y2, r)
			local x = x1 * (1-r) + x2 * r
			local y = y1 * (1-r) + y2 * r
			return x, y
		end
		
		local tlx, tly = love.graphics.getWidth() - 210, 10
		local lx, ly = interpolate(x - cx, y - cy, tlx, tly + 20, transition)
		
		love.graphics.setInvertedStencil(stencilFunc)
		love.graphics.setColor(0, 255 * transition, 255 * (1 - transition))
		love.graphics.line(x - cx, y - cy, lx, ly)
		--love.graphics.line(tlx, tly + 20, lx, ly)
		love.graphics.setInvertedStencil()
		
		local offset = tly + 10
		
		local font = love.graphics.getFont()
		local fontHeight = font:getHeight()
		
		local boxWidth = 200 - 20
		
		local function showText(msg, leftOffset, topOffset)
			-- Offset the offset (...wait, what?)
			if topOffset then
				offset = offset + topOffset
			end
			
			leftOffset = leftOffset or 0
			
			love.graphics.setColor(0, 255, 0, 255 * transition)
			love.graphics.printf(msg, tlx + 10 + leftOffset, offset,
					boxWidth - leftOffset)
			
			local width, lines = font:getWrap(msg, boxWidth)
			offset = offset + fontHeight * lines + 5
		end
		
		showText("Object ID: " .. object.id)
		showText("Blueprint: " .. object.blueprint.name)
		
		if object.body then
			-- Speed in km/h
			local vx, vy = object.body:getLinearVelocity()
			local speed = math.sqrt(vx*vx + vy*vy) / pixelsPerMeter
			local kph = math.floor(speed * 3.6)
			
			-- The mass is supposedly in kg, but... until it actually is, we'll
			-- use "mass units" instead, to not immediately break immersion.
			-- ToDo: get the mass of the fixture instead
			local mass = string.format("%.2f", object.body:getMass())
			
			showText("Mass: " .. mass .. "mu")
			showText("Speed: " .. kph .. "km/h")
		else
			-- ToDo: for top-mounted modules, get the velocity of the hardpoint
			-- and the mass of the fixture.
		end
		
		showText("Description: An object that was apparently materialized in this world at some point in time.")
		
		infoBox = { showText = showText }
		
		if object.blueprint.showInfo then
			object.blueprint.showInfo(object, infoBox)
		end
		
		if object.blueprint.topHardpoints then
			showText("Upgrade hardpoints:")
			for i, topHp in ipairs(object.blueprint.topHardpoints) do
				local msg = "#" .. i .. " - "
				if object.connectedObjects[topHp] then
					msg = msg .. object.connectedObjects[topHp].blueprint.name
				else
					msg = msg .. "[empty]"
				end
				showText(msg, 5, -3)
			end
		end
		
		-- The box is actually above the text... but no one will notice, right?
		love.graphics.setColor(0, 255, 0, 32 * transition)
		love.graphics.rectangle("fill", tlx + 0.5, tly + 0.5, 200, offset - 8)
		love.graphics.setColor(0, 255, 0, 255 * transition)
		love.graphics.rectangle("line", tlx + 0.5, tly + 0.5, 200, offset - 8)
	end
	
	-- Show currently selected object
	if selectedObject then
		showObjectInfo(selectedObject, selectedObjectGuiTransition)
	end
	
	if selectedObjectGuiLast then
		showObjectInfo(selectedObjectGuiLast, 1 - selectedObjectGuiTransition)
	end
end

function love.mousepressed(mx, my, button)
	local width, height = love.graphics.getWidth(), love.graphics.getHeight()
	
	local cx, cy = getCameraToWorldPoint(mx, my)
	
	if (button == "wd" or button == "wu") and activeBlueprint then
		currentConnector = currentConnector + (button == "wd" and 1 or -1)
		currentConnector = minMax(currentConnector, 1, #(activeBlueprint.connectors))
	end
	
	if button == "l" and objects.vehicle then
		-- Destroy object (if it is destroyable)
		if love.keyboard.isDown("r") then
			local mx, my = getCameraToWorldPoint(love.mouse.getPosition())
			
			local obj, distance = findNearestObject(mx, my, 50, function(obj)
				return obj.blueprint ~= templates.vehicle
			end, true)
			
			if obj and obj.partOfBlueprint then
				removeFromCustomBlueprint(obj)
			elseif obj and obj.blueprint.destroy and
					distanceFromVehicle(obj:getPosition()) < 300 then
				sendCall.disintegrateObject(obj.id)
			end
			
			return
		end
		
		for i = 1, 9 do
			local boxSize = 30
			local x, y = 50 + (i-1) * (boxSize + 10), height - 50 - boxSize
			
			if mx >= x and mx <= x + boxSize and my >= y and my <= y + boxSize then
				activeBlueprint = blueprintBar[i]
				activeBlueprintButton = i
				activeBlueprintGuiTransition = 0
				currentConnector = 1
				return
			end
		end
		
		if activeBlueprint and distanceFromVehicle(cx, cy) < 200 then
			if activeBlueprint.isCustomBlueprint then
				sendMaterializeCustomBlueprint(activeBlueprint, cx, cy, objects.vehicle.body:getAngle())
				
				activeBlueprint = nil
			elseif activeBlueprint.build then
				-- Materialize stuff
				
				-- Check for the nearest hard-point to connect to, if any
				local obj, hardpoint, hardpointNr = findNearestHardpoint(cx, cy)
				
				if hardpoint and activeBlueprint.connectors then
					guiMaterializeAndConnectObject(activeBlueprint, obj,
							hardpoint, hardpointNr,
							activeBlueprint.connectors[currentConnector], currentConnector)

					--[[
					if designingCustomBlueprint then
						local function bpObjGetPosition(self)
							return self.x, self.y
						end
						
						local bpObj = {
							blueprint = activeBlueprint,
							type = activeBlueprint.name,
							toObj = obj, toObjId = obj.id or -obj.blueprintIndex,
							hardpoint = hardpoint, hardpointNr = hardpointNr,
							connector = activeBlueprint.connectors[currentConnector],
							connectorNr = currentConnector,
							partOfBlueprint = true,
							blueprintIndex = #customBlueprint + 1,
							
							getPosition = bpObjGetPosition,
						}
						
						table.insert(customBlueprint, bpObj)
					else
						sendToServer("materializeAndConnectObject",
								{type = activeBlueprint.name, to = obj.id,
								 hardpoint = hardpointNr,
								 connector = currentConnector})
					end
					--]]
				else
					local cr = objects.vehicle.body:getAngle()
					-- The math.floor() calls should give smaller string
					-- representations for the object's coordinates and angle
					local x, y, r = math.floor(cx), math.floor(cy),
							math.floor(cr * 100) / 100
					
					guiMaterializeObject(activeBlueprint, x, y, r)
					
					--[[
					if designingCustomBlueprint then
						local function bpObjGetPosition(self)
							return self.x, self.y
						end
						
						local bpObj = {
							blueprint = activeBlueprint,
							type = activeBlueprint.name,
							x = x, y = y, r = r,
							partOfBlueprint = true,
							blueprintIndex = #customBlueprint + 1,
							
							getPosition = bpObjGetPosition,
						}
						table.insert(customBlueprint, bpObj)
					else
						sendToServer("materializeObject",
								{type = activeBlueprint.name, x = x, y = y, r = r})
					end
					--]]
				end
				
				activeBlueprint = nil
			elseif activeBlueprint.mountOn then
				-- Materialize a top-mounted upgrade module
				
				local obj, topHp, topHpNr = findNearestTopHardpoint(cx, cy)
				
				if obj and not obj.connectedObjects[topHp] then
					sendCall.mountObject(activeBlueprint.name, obj.id, topHpNr)
					
					activeBlueprint = nil
				end
			else
				-- Not buildable... do nothing.
			end
			
			return
		end
		
		if love.mouse.isDown("r") then
			-- Shoot laser
			sendCall.fireLaser()
		else
			-- Select stuff
			local obj = findNearestObject(cx, cy, 30, function (obj)
				return not obj.mountedOn
			end)
			
			selectedObjectGuiLast = selectedObject
			selectedObjectGuiTransition = 0
			keepSelectedObjectVisible = 1
			
			selectedObject = obj
		end
	end
end

function guiMaterializeObject(blueprint, x, y, r)
	if designingCustomBlueprint then
		local function bpObjGetPosition(self)
			return self.x, self.y
		end
		
		local bpObj = {
			blueprint = blueprint,
			type = blueprint.name,
			x = x, y = y, r = r,
			partOfBlueprint = true,
			blueprintIndex = #customBlueprint + 1,
			
			getPosition = bpObjGetPosition,
		}
		table.insert(customBlueprint, bpObj)
	else
		sendToServer("materializeObject",
				{type = blueprint.name, x = x, y = y, r = r})
	end
end

function guiMaterializeAndConnectObject(blueprint, toObj,
		hardpoint, hardpointNr, connector, connectorNr)
	if designingCustomBlueprint then
		local function bpObjGetPosition(self)
			return self.x, self.y
		end
		
		local bpObj = {
			blueprint = blueprint,
			type = blueprint.name,
			toObj = toObj, toObjId = toObj.id or -toObj.blueprintIndex,
			hardpoint = hardpoint, hardpointNr = hardpointNr,
			connector = connector,
			connectorNr = connectorNr,
			partOfBlueprint = true,
			blueprintIndex = #customBlueprint + 1,
			
			getPosition = bpObjGetPosition,
		}
		
		table.insert(customBlueprint, bpObj)
	else
		sendToServer("materializeAndConnectObject",
				{type = blueprint.name, to = toObj.id,
				 hardpoint = hardpointNr,
				 connector = connectorNr})
	end
end

-- ToDo: low-priority, but... maybe make blueprint objects be real objects,
-- with the help of a no-physical-form chameleon template?
function removeFromCustomBlueprint(doomedObj)
	for i, bpObj in ipairs(customBlueprint) do
		-- Remove all references
		if bpObj.toObj == doomedObj then
			bpObj.toObj = nil
		end
		
		-- Update blueprint object ids
		if doomedObj.partOfBlueprint and i > doomedObj.blueprintIndex then
			bpObj.blueprintIndex = bpObj.blueprintIndex - 1
		end
	end
	
	if doomedObj.partOfBlueprint then
		table.remove(customBlueprint, doomedObj.blueprintIndex)
	end
end

function addConnectedObjectToBlueprint(fromVehicle, type, toObj, hardpointNr, connectorNr)
	fromVehicle.customBlueprint = fromVehicle.customBlueprint or { }
	
	local bpObj = {
		partOfBlueprint = true,
		blueprintIndex = #fromVehicle.customBlueprint + 1
	}
	
	bpObj.type = type
	bpObj.blueprint = templates[type]
	
	-- References to previous objects in the custom blueprint have negative ids
	if toObj < 0 then
		bpObj.toObj = fromVehicle.customBlueprint[-toObj]
	else
		bpObj.toObj = objectById[toObj]
	end
	
	if not bpObj.blueprint or not bpObj.toObj then
		-- ToDo: improve error system with 'return nil, message'
		print("Malformed addToBlueprint message.")
		return
	end
	
	bpObj.hardpointNr = hardpointNr
	bpObj.connectorNr = connectorNr
	
	bpObj.hardpoint = bpObj.toObj.blueprint.sideHardpoints[hardpointNr]
	bpObj.connector = bpObj.blueprint.connectors[connectorNr]
	
	if not bpObj.hardpoint or not bpObj.connector then
		print("Malformed addToBlueprint message (e2).")
		return
	end
	
	table.insert(fromVehicle.customBlueprint, bpObj)
end

function addObjectToBlueprint(fromVehicle, type, x, y, r)
	fromVehicle.customBlueprint = fromVehicle.customBlueprint or { }
	
	local bpObj = {
		partOfBlueprint = true,
		blueprintIndex = #fromVehicle.customBlueprint + 1
	}
	
	bpObj.type = type
	bpObj.blueprint = templates[type]
	
	if not bpObj.blueprint then
		-- ToDo: improve error system with 'return nil, message'
		print("Malformed addToBlueprint message (e1).")
		return
	end
	
	bpObj.x, bpObj.y, bpObj.r = x, y, r
	
	table.insert(fromVehicle.customBlueprint, bpObj)
end

function materializeCustomBlueprint(fromVehicle, cx, cy, cr)
	if not fromVehicle.customBlueprint then
		return
	end
	
	for i, bpObj in ipairs(fromVehicle.customBlueprint) do
		if not bpObj.toObj then
			local x, y = localToWorld(cx, cy, cr, bpObj.x, bpObj.y)
			local r = cr + bpObj.r
			
			local obj = materializeObject(fromVehicle, bpObj.type, x, y, r)
			bpObj.realObject = obj
		else
			local connectTo = bpObj.toObj
			if connectTo.partOfBlueprint then
				connectTo = connectTo.realObject
			end
			
			-- function materializeAndConnectObject(fromVehicle, blueprintName,
			--		objId, hardpointNr, connectorNr)
			if connectTo then
				local obj = materializeAndConnectObject(fromVehicle, bpObj.type,
						connectTo.id, bpObj.hardpointNr, bpObj.connectorNr)
				bpObj.realObject = obj
			end
		end
	end
	
	fromVehicle.customBlueprint = nil
end

function addExplosion(x, y)
	local explosion = love.graphics.newParticleSystem(whitePixel, 30)
	explosion:setColors(255, 255, 0, 255,   255, 0, 0, 255)
	explosion:setPosition(x, y)
	explosion:setLifetime(0.1)
	explosion:setEmissionRate(300) -- 30 particles over 0.1 seconds
	explosion:setParticleLife(0.5, 1)
	explosion:setSpread(2 * math.pi)
	explosion:setSpeed(50, 100)
	explosion:setSizes(2, 2, 0)
	explosion:start()
	
	ephemeralParticleSystems[explosion] = true
end

function addChainLinkDispersion(x, y)
	local sparks = love.graphics.newParticleSystem(bluePixel, 5)
	
	sparks:setColors(255, 255, 255, 255,   255, 255, 255, 0)
	sparks:setPosition(x, y)
	sparks:setLifetime(0.1)
	sparks:setEmissionRate(50)
	sparks:setParticleLife(0.5, 1)
	sparks:setSpread(2 * math.pi)
	sparks:setSpeed(10, 40)
	sparks:setSizes(2, 2, 0)
	sparks:start()
	
	ephemeralParticleSystems[sparks] = true
end

function createLaserBeam(fromObject, offset, length, beamLife, instant)
	local x, y = fromObject.body:getWorldPoint(length, 0)
	local vx, vy = fromObject.body:getWorldPoint(offset, 0)
	local hx, hy, nx, ny, obj
	
	local function rayCastCallback(fixture, pointX, pointY, normalX, normalY, fraction)
		hx, hy = pointX, pointY
		nx, ny = normalX, normalY
		obj = fixture:getUserData()
		return fraction
	end
	
	world:rayCast(vx, vy, x, y, rayCastCallback)
	
	if not hx or not hy then
		hx, hy = x, y
		nx, ny = 1, 0
	end
	
	local life = beamLife
	beam = { coords = { vx, vy, hx, hy }, life = life, maxLife = life, instant = instant }
	laserBeams[beam] = true
	
	-- Do some damage
	if obj and obj.blueprint and obj.blueprint.damage then
		local ex, ey = obj.body:getPosition()
		obj.blueprint.damage(obj)
	end
	
	-- Add some particles at the point of impact
	local sparks = love.graphics.newParticleSystem(bluePixel, 10)
	sparks:setPosition(hx, hy)
	sparks:setLifetime(0.1)
	sparks:setEmissionRate(100) -- 10 particles over 0.1 seconds
	sparks:setParticleLife(0.3, 0.5)
	sparks:setSpread(math.pi * 0.80)
	sparks:setDirection(math.atan2(ny, nx))
	sparks:setSpeed(100)
	sparks:setSizes(2, 2, 0)
	sparks:start()
	
	ephemeralParticleSystems[sparks] = true
end

function fireLaser(fromVehicle)
	createLaserBeam(fromVehicle, 9, 1000, 0.5)
end

function fireHook(fromObject)
	if fromObject.hooked then
		return
	end
	
	fromObject.hooked = true
	
	-- Create the hook in front of the vehicle
	local x, y = fromObject.body:getWorldPoint(10+6, 0)
	local r = fromObject.body:getAngle()
	local hook = registerObject(templates.hook.build(x, y, r))
	hook.fromObject = fromObject
	
	fromObject:callWhenDestroyed(function (self)
		hook.fromObject = nil
	end)
	
	-- Add an impulse to make it fly
	local speed = 100
	local ix, iy = fromObject.body:getWorldPoint(speed, 0)
	hook.body:applyLinearImpulse(ix - x, iy - y)
	
	-- And one in the opposite direction to the thing that fired it
	fromObject.body:applyLinearImpulse(x - ix, y - iy)
end

function fireVehicleHook(fromVehicle)
	fireHook(fromVehicle)
end

function radialSignal(fromVehicle)
	local radius = 200
	local x, y = fromVehicle.body:getPosition()
	
	-- Call the 'activate' method of all nearby objects
	-- ToDo: find a way to fix this ugly sorting hack
	local sortedObjects = { }
	for obj in pairs(dynamicObjects) do
		local dist = distance(x, y, obj:getPosition())
		
		if obj.blueprint.activate and dist < radius then
			table.insert(sortedObjects, obj)
		end
	end
	table.sort(sortedObjects, function (a, b) return a.id < b.id end)
	for i, obj in ipairs(sortedObjects) do
		if not obj.destroyed then
			obj.blueprint.activate(obj)
		end
	end
end


dynamicObjects = { }
objectById = { }

nextId = 1

function registerObject(object, id)
	if id == 0 then
		print("Attempted to register object with id '0'! (oh, and attempt ignored)")
	end
	
	id = id or nextId
	
	object.id = id
	if id > 0 then
		dynamicObjects[object] = true
	end
	objectById[object.id] = object
	
	if id >= nextId then
		nextId = id + 1
	end
	
	object.callWhenDestroyed = object.callWhenDestroyed or function(self, func)
		-- Walls never get destroyed... so don't do anything
		if not self.wall then
			self.destroyCallbacks = self.destroyCallbacks or { }
			table.insert(self.destroyCallbacks, func)
		end
	end
	
	return object
end

function materializeObject(fromVehicle, blueprintName, x, y, r)
	local blueprint = templates[blueprintName]
	
	if not blueprint then
		print("Trying to materialize from unknown blueprint (" ..
				blueprintName .. ")")
		return
	end
	
	local o = registerObject(blueprint.build(x, y, r))
	
	return o
end

function connectObjectTo(obj, toObj, hardpointNr, connectorNr, blueprintName)
	local blueprint = obj and obj.blueprint or templates[blueprintName]
	
	if blueprintName and not blueprint then
		print("Trying to materialize from unknown blueprint (" ..
				blueprintName .. ")")
		return
	end
	
	local connector = blueprint.connectors[connectorNr]
	local hardpoint = toObj.blueprint.sideHardpoints[hardpointNr]
	
	if not connector or not hardpoint then
		print("Trying to connect with invalid connector/hardpoint (" ..
				blueprintName .. ", " .. tostring(connectorNr) .. "/" ..
				tostring(hardpointNr) .. ")")
		return
	end
	
	local x, y = toObj.body:getWorldPoint(unpack(hardpoint))
	local angle = toObj.body:getAngle() + hardpoint.angle - connector.angle - math.pi
	
	if blueprintName then
		obj = materializeObject(fromVehicle, blueprintName, x, y, angle)
		if not obj then
			return
		end
	else
		obj.body:setPosition(x, y)
		obj.body:setAngle(angle)
	end

	-- A bit of voodoo to avoid some sin/cos calls in Lua
	local x2, y2 = unpack(connector)
	obj.body:setPosition(obj.body:getWorldPoint(-x2, -y2))
	
	local j = love.physics.newWeldJoint(obj.body, toObj.body, x, y, false)
	
	obj.connectedTo = {
		toObject = toObj,
		connectorNr = connectorNr,
		hardpointNr = hardpointNr
	}
	
	toObj:callWhenDestroyed(function (self)
		obj.connectedTo = nil
	end)
	
	-- ToDo: register joint
	
	return obj
end

-- Similar to the above, but materialize on the right spot
function materializeAndConnectObject(fromVehicle, blueprintName, objId, hardpointNr, connectorNr)
	local obj = objectById[objId]
	
	if not obj then
		print("Trying to materialize on unknown object id (" .. objId .. ")")
		return
	end
	
	return connectObjectTo(nil, obj, hardpointNr, connectorNr, blueprintName)
end

function connectObjectOn(object, onObject, topHpNr)
	local topHp = onObject.blueprint.topHardpoints[topHpNr]
	
	object.blueprint.remountOn(object, onObject, topHp)
	object.mountedOn = onObject
	object.connectedTo = {toObject = onObject, hardpointNr = topHpNr, connectorNr = 0}
	onObject.connectedObjects[topHp] = object
	
	object:callWhenDestroyed(function (self)
		if onObject.connectedObjects[topHp] == object then
			onObject.connectedObjects[topHp] = nil
		end
	end)
end

function mountObject(fromVehicle, blueprintName, objId, topHpNr)
	local blueprint = templates[blueprintName]
	local object = objectById[objId]
	local topHp = object and object.blueprint.topHardpoints and
			object.blueprint.topHardpoints[topHpNr]
	
	if not blueprint or not blueprint.mountOn then
		print("Attempting to mount with an invalid blueprint (" ..
				blueprintName .. ")")
		return
	end
	
	if not object then
		print("Attempting to mount on invalid object (" .. objId .. ")")
		return
	end
	
	if not topHp then
		print("Attempting to mount on invalid top-hardpoint (" .. topHpNr .. ")")
		return
	end
	
	-- ToDo: check distance (ditto for other materialization functions)
	
	local module = registerObject(blueprint.mountOn(object, topHp))
	module.mountedOn = object
	module.connectedTo = {toObject = object, hardpointNr = topHpNr, connectorNr = 0}
	object.connectedObjects[topHp] = module
	
	module:callWhenDestroyed(function (self)
		if object.connectedObjects[topHp] == module then
			object.connectedObjects[topHp] = nil
		end
	end)
end

function disintegrateObject(fromVehicle, objId)
	local obj = objectById[objId]
	
	if not obj then
		print("Trying to disintegrate with an unknown object id (" .. objId .. ")")
		return
	end
	
	explodeObject(obj)
end

function explodeObject(obj)
	addExplosion(obj:getPosition())
	destroyObject(obj)
end

function destroyObject(obj)
	-- Things welded to the side are just unwelded, but top-mounted modules
	-- must be completely destroyed as well.
	if obj.connectedObjects then
		for i, topHp in ipairs(obj.blueprint.topHardpoints or { }) do
			local module = obj.connectedObjects[topHp]
			
			if module then
				destroyObject(module)
			end
		end
	end
	
	-- Some custom blueprint objects might be connected to this
	removeFromCustomBlueprint(obj)
	
	obj.blueprint.destroy(obj)
	dynamicObjects[obj] = nil
	objectById[obj.id] = nil
	if obj.player then
		players[obj] = nil
	end
	
	if obj.destroyCallbacks then
		for i, destroyCallback in ipairs(obj.destroyCallbacks) do
			destroyCallback(obj)
		end
	end
	
	obj.destroyed = true
	
	if selectedObjectGuiLast == obj then
		selectedObjectGuiLast = nil
	end
	if selectedObject == obj then
		selectedObject = nil
	end
end
