
-- ToDo when the server is threaded:
--require("love.filesystem")
--love.filesystem.load("settings.lua")()

local socket = require("socket")

local serverSocket
local clients = {}
local clientBySocket = {}

local serverTimeStep = 1 / 60
local currentTick = 1
local currentTickTime

local tickReports = {}

local eol = settings.serverEOL or "\n"


function openServer()
	local s, err = socket.bind("*", settings.serverPort)
	
	serverSocket = s
	
	if not serverSocket then
		io.stderr:write("[server - socket.bind()] - " .. err .. "\n")
		return false
	end
	
	serverSocket:settimeout(0)
	
	return true
end

function getTime()
	return love.timer.getMicroTime()
	--return socket.gettime()
end

function checkServer()
	mainLoop(0, true)
end

function mainLoop(timeout, singleIteration)
	if not serverSocket then
		return
	end

	if not currentTickTime then
		currentTickTime = getTime()
	end
	
	while true do
		local checkList = {serverSocket}
		for client in pairs(clients) do
			table.insert(checkList, client.socket)
		end
		
		--io.stderr:write("Selecting...\n")
		local readableSockets, writableSockets, err = socket.select(checkList, nil, timeout)
		--io.stderr:write("Selected!\n")
		
		if err == "timeout" then
			local newTickTime = getTime()
			while currentTickTime + serverTimeStep <= newTickTime do
				currentTickTime = currentTickTime + serverTimeStep
				currentTick = currentTick + 1
				handleTick(currentTick)
			end
			
			if singleIteration then
				return
			end
		else
			for i, s in ipairs(readableSockets) do
				if s == serverSocket then
					local newClient, err = serverSocket:accept()
					
					if newClient then
						newClient:settimeout(0)
						newClient:setoption("tcp-nodelay", true)
						local client = createClient(newClient)
						client.partialMsg = ""
					elseif err ~= "timeout" then
						io.stderr:write("[server - server:accept()] - " .. err .. "\n")
					end
				else
					local client = clientBySocket[s]
					local msg, err, partialMsg = s:receive("*l")
					
					if not msg then
						if partialMsg then
							client.partialMsg = client.partialMsg .. partialMsg
						end
						
						if err ~= "timeout" then
							if err ~= "closed" then
								io.stderr:write("[server - client:read()] - " .. err .. "\n")
							end
							
							closeClient(client)
						end
					else
						handleClient(client, client.partialMsg .. msg)
						client.partialMsg = ""
					end
				end
			end
		end
	end
end

local sendGameState = false

local nextName = 1

function createClient(s)
	local client = {}
	
	client.socket = s
	
	clients[client] = true
	clientBySocket[s] = client
	client.peerName = s:getpeername()
	
	print("New client created (" .. client.peerName .. ").")
	
	client.playerName = "P" .. nextName
	nextName = nextName + 1
	
	client.prefix = "player<" .. client.playerName .. ">: "
	
	sendToClient(client, "server: Welcome! You are: name<" .. client.playerName .. ">" .. eol)
	
	-- Fetch and send the game state from the master client
	sendGameState = true
	
	return client
end

function closeClient(client)
	print("Client closed (" .. client.peerName .. ").")
	
	buffer = buffer .. "server: player-disconnected name<" .. client.playerName .. ">" .. eol
	
	clients[client] = nil
	clientBySocket[client.socket] = nil
end

function handleClient(client, msg)
	local tick = string.match(msg, "ping (%d+)")
	
	if tick then
		sendToClient(client, "server: pingReply " .. tick .. eol)
	else
		buffer = buffer .. client.prefix .. msg .. eol
	end
end

buffer = ""

function handleTick(tick)
	if tick % 100 == 0 then
		--print("Tick.")
	elseif tick % 100 == 50 then
		--print("Tock.")
	end
	
	if tick % 60 == 0 then
		buffer = buffer .. "server: tick " .. tick .. " report request" .. eol
	end
	
	-- ToDo: ask for and get the state via TCP/love.thread.demand() instead
	local state = sendGameState and ("server: state-broadcast - [" ..
			saveGameState() .. "]" .. eol) or ""
	sendGameState = false
	
	sendToAll(state .. buffer .. ">" .. eol)
	buffer = ""
end

function sendToClient(client, msg)
	-- Maybe switch to blocking mode while sending?
	-- ToDo: Nope, it's awful. Need to implement proper write buffering.	
	client.socket:settimeout(nil)
	
	local lastByte, err = client.socket:send(msg)
	
	if err then
		io.stderr:write("[server - client:write()] - " .. err .. "\n")
	end
	
	client.socket:settimeout(0)
end

function sendToAll(msg)
	for c in pairs(clients) do
		sendToClient(c, msg)
	end
end

function startServer()
	openServer()
	--mainLoop(0.003)
end

