
vectorMetatable =
{
	__add = function(v1, v2)
		return vector{x = v1.x + v2.x, y = v1.y + v2.y}
	end,

	__mul = function(v1, v2)
		local scalar = tonumber(v2)
		
		if scalar then
			return {x = v1.x * scalar, y = v1.y * scalar}
		end
		
		return nil -- implement vector multiplication, if needed
	end,
	
	__index = function(v, key)
		if key == "left" then
			return vector{x = v.y, y = -v.x}
		elseif key == "right" then
			return vector{x = -v.y, y = v.x}
		elseif key == "rotate" then
			return function(angle)
				local r = math.atan2(v.y, v.x)
				return vector{x = math.cos(r + angle), y = math.sin(r + angle)}
			end
		end
	end
}

function vector(v)
	return setmetatable(v, vectorMetatable)
end

north = vector{x = 0,  y = -1}
east  = vector{x = 1,  y = 0}
south = vector{x = 0,  y = 1}
west  = vector{x = -1, y = 0}

--[[
north.left, north.right = west, east
east.left, east.right = north, south
south.left, south.right = east, west
west.left, west.right = south, north
--]]

curPos = vector{x = 0, y = 0}
curAngle = north
halfWidth = 25


function moveForward(length)
	curPos = curPos + curAngle * length
end

function turnLeft()
	curPos = curPos + curAngle.left * (halfWidth * 1.01) + curAngle * (halfWidth * 1.01)
	curAngle = curAngle.left
end

function turn(angle)
	curAngle = curAngle.rotate(angle)
end

function turnTo(angle)
	curAngle = angle
end

function turnSlightlyLeft()
	
end

oldWallPos = nil
oldWallAngle = nil

function startWall(setPos, setAngle)
	oldWallPos = setPos
	oldWallAngle = setAngle
	curPos = setPos
	curAngle = setAngle
end

function makeWall()
	local p1 = oldWallPos + oldWallAngle.left * halfWidth
	local p2 = oldWallPos + oldWallAngle.right * halfWidth
	
	local p4 = curPos + curAngle.left * halfWidth
	local p3 = curPos + curAngle.right * halfWidth
	
	table.insert(walls, { x1 = p1.x, y1 = p1.y, x2 = p2.x, y2 = p2.y,
	                      x3 = p3.x, y3 = p3.y, x4 = p4.x, y4 = p4.y })
	
	oldWallPos = curPos
	oldWallAngle = curAngle
end

function makeDoubleWall(roadHalfWidth)
	local p1, p2, p3, p4, p5, p6, p7, p8
	
	-- Left wall
	p1 = oldWallPos + oldWallAngle.left * (halfWidth * 2 + roadHalfWidth)
	p2 = oldWallPos + oldWallAngle.left * (halfWidth * 0 + roadHalfWidth)
	
	p4 = curPos + curAngle.left * (halfWidth * 2 + roadHalfWidth)
	p3 = curPos + curAngle.left * (halfWidth * 0 + roadHalfWidth)
	
	table.insert(walls, { x1 = p1.x, y1 = p1.y, x2 = p2.x, y2 = p2.y,
	                      x3 = p3.x, y3 = p3.y, x4 = p4.x, y4 = p4.y })
	
	-- Right wall
	p5 = oldWallPos + oldWallAngle.right * (halfWidth * 2 + roadHalfWidth)
	p6 = oldWallPos + oldWallAngle.right * (halfWidth * 0 + roadHalfWidth)
	
	p8 = curPos + curAngle.right * (halfWidth * 2 + roadHalfWidth)
	p7 = curPos + curAngle.right * (halfWidth * 0 + roadHalfWidth)
	
	table.insert(walls, { x1 = p5.x, y1 = p5.y, x2 = p6.x, y2 = p6.y,
	                      x3 = p7.x, y3 = p7.y, x4 = p8.x, y4 = p8.y })
	
	-- Ground between
	-- p2, p3, p7, p6
	table.insert(ground, { x1 = p2.x, y1 = p2.y, x2 = p3.x, y2 = p3.y,
	                       x3 = p7.x, y3 = p7.y, x4 = p6.x, y4 = p6.y })
	
	oldWallPos = curPos
	oldWallAngle = curAngle
end
